#include "../include/bigint.hpp"

CBigInt::CBigInt() : m_negative(false) {}
CBigInt::CBigInt( int64_t parent ) : m_negative(parent < 0) {
        m_data.reserve( CHUNK_BYTES / sizeof(parent) );
        uint64_t par = std::abs(parent);
        if (CHUNK_BITS == 64) {
            m_data.push_back(par);
            return;
        }

        while (par) {
            m_data.push_back(par);
            par >>= CHUNK_BITS;
        }
    }

CBigInt::CBigInt( const CBigInt &oth ) : m_data(oth.m_data), m_negative(oth.m_negative) {}
CBigInt::~CBigInt() {}

std::string CBigInt::toString() const {
    std::string retStr;
    if (m_negative)
        retStr = '-';
    
    if (m_data.size() <= 1) {
        // If the number fits within a single word, use sts conversion
        uint64_t ret = 0;
        for (uint64_t i = 0; i < sizeof(ret) / CHUNK_BYTES && i < m_data.size(); ++i)
            ret += m_data[i] << (8 * i);
        return retStr + std::to_string(ret);
    }

    std::unique_ptr<CBigInt> divNum (new CBigInt(*this));
    CBigInt ten(10);
    while (!divNum->isZero()) {
        auto tmp = divNum->divmod(&ten);
        divNum.reset(tmp.first);
        retStr += std::to_string(tmp.second->m_data[0]);
        delete tmp.second;
    }

    // Reverse the string representation of the number
    for (uint64_t i = 0; i < retStr.size() / 2; i++)
        std::swap(retStr[i], retStr[retStr.size() - i - 1]);
    return retStr;
}

void CBigInt::debugPrint() {
        uint64_t tmp = 0;
        if (m_negative)
            std::cout << '-';
        for (uint64_t i = 0; i < sizeof(tmp) / CHUNK_BYTES && i < m_data.size(); ++i)
            tmp += (uint64_t)m_data[i] << (CHUNK_BITS * i);
        std::cout << tmp << '\n';
    }

int64_t CBigInt::debugDump() {
    int64_t tmp = 0;
    for (uint64_t i = 0; i < sizeof(tmp) / CHUNK_BYTES && i < m_data.size(); ++i)
        tmp += (uint64_t)m_data[i] << (CHUNK_BITS * i);
    if (m_negative)
        tmp *= -1;
    return tmp;
}

CBigInt *CBigInt::add( const CBigInt *oth ) const {
    if (isZero())
        return new CBigInt(*oth);
    if (oth->isZero())
        return new CBigInt(*this);

    if (!(m_negative ^ oth->m_negative)) {

        CBigInt *ret = new CBigInt();
        ret->m_data.reserve( std::max(m_data.size() + 1, oth->m_data.size() + 1) );
        bool carry = 0;

        for (uint64_t i = 0; i < std::max(m_data.size(), oth->m_data.size()); ++i) {
            CHUNK_TYPE tmp = ( i < m_data.size() ? m_data[i] : 0 ) + ( i < oth->m_data.size() ? oth->m_data[i] : 0 ) + carry;
            carry = (tmp < ((i < m_data.size()) ? m_data[i] : oth->m_data[i]));
            ret->m_data.push_back(tmp);
        }
        if (carry)
            ret->m_data.push_back(1);

        ret->m_negative = m_negative;
        return ret;
    }

    // If the numbers have different signs
    int comp = cmp(oth);
    CBigInt *bigger  = (CBigInt *)((comp >= 0) * (size_t)this + (comp < 0)   * (size_t) oth);
    CBigInt *smaller = (CBigInt *)((comp >= 0) * (size_t)oth + (comp < 0)   * (size_t) this);
    bool tmpB = bigger->m_negative; 
    bigger->m_negative = false;
    bool tmpS = smaller->m_negative;
    smaller->m_negative = false;

    CBigInt *ret = bigger->sub(smaller);
    ret->m_negative = tmpB;

    bigger->m_negative = tmpB;
    smaller->m_negative = tmpS;

    return ret;
}

CBigInt *CBigInt::sub( const CBigInt *oth ) const {
    int comp = cmp(oth);
    CBigInt *bigger  = (CBigInt *)((comp >= 0) * (size_t)this + (comp < 0)   * (size_t) oth);
    CBigInt *smaller = (CBigInt *)((comp >= 0) * (size_t)oth + (comp < 0)   * (size_t) this);

    if (!(m_negative ^ oth->m_negative)) {

        CBigInt *ret = new CBigInt();
        ret->m_data.reserve( bigger->m_data.size() );
        bool borrow = 0;

        for (; smaller->m_data.size() < bigger->m_data.size(); smaller->m_data.push_back(0));

        for (uint64_t i = 0; i < smaller->m_data.size(); ++i) {
            CHUNK_TYPE tmp = bigger->m_data[i];
            if (borrow) {
                if (bigger->m_data[i] == 0) {
                    tmp = CHUNK_MAX;
                } else {
                    tmp = bigger->m_data[i] - 1;
                    borrow = false;
                }
            }

            borrow |= bigger->m_data[i] < smaller->m_data[i];
            tmp -= smaller->m_data[i];
            ret->m_data.push_back(tmp);
        }

        ret->m_negative = m_negative;
        if (this == smaller)
            ret->m_negative = !ret->m_negative;
        return ret;
    }

    // If the numbers have different signs
    bool tmpB = bigger->m_negative; 
    bigger->m_negative = false;
    bool tmpS = smaller->m_negative;
    smaller->m_negative = false;

    CBigInt *ret = bigger->add(smaller);
    ret->m_negative = tmpB;

    bigger->m_negative = tmpB;
    smaller->m_negative = tmpS;

    return ret;
}

CBigInt *CBigInt::mul( const CBigInt *oth ) const {
    std::unique_ptr<CBigInt> ret( new CBigInt() );
    std::unique_ptr<CBigInt> shifted( new CBigInt(*this) );

    for (uint64_t i = 0; i < (oth->m_data.size() * CHUNK_BITS); ++i) {
        if (oth->getBit(i)) {
            ret->addDestructive( shifted.get() );
        }
        shifted->pushLeft(0);
    }
    ret->m_negative = m_negative ^ oth->m_negative;
    return ret.release();
}

CBigInt *CBigInt::div( const CBigInt *oth ) const {
    auto ret = divmod(oth);
    delete ret.second;
    return ret.first;
}

CBigInt *CBigInt::mod( const CBigInt *oth ) const {
    auto ret = divmod(oth);
    delete ret.first;
    return ret.second;
}

CBigInt *CBigInt::gcd( const CBigInt *oth ) const {
    auto ret = rea(oth);
    delete ret.second;
    return ret.first;
}

CBigInt *CBigInt::lcm( const CBigInt *oth ) const {
    auto ret = rea(oth);
    delete ret.first;
    return ret.second;
}

CBigInt *CBigInt::shf( int64_t shift ) const {
    CBigInt *ret = new CBigInt();
    if (shift == 0) {
        delete ret;
        ret = new CBigInt(*this);
    }

    if (shift < 0) {
        throw std::logic_error("invalid bitshift");
    }

    if (shift > 0) {
        size_t bitOffset = shift % CHUNK_BITS;
        size_t chunkOffset = shift / CHUNK_BITS;
        size_t numNewChunks = m_data.size() + ((shift + CHUNK_BITS - 1) / CHUNK_BITS);

        ret->m_data = std::vector<CHUNK_TYPE>( chunkOffset, 0 );
        ret->m_data.reserve( numNewChunks );

        CHUNK_TYPE carry = 0;
        for (uint64_t i = 0; i < m_data.size(); ++i) {
            ret->m_data.push_back(carry | (m_data[i] << bitOffset));
            carry = m_data[i] >> (CHUNK_BITS - bitOffset);
        }
        if (carry)
            ret->m_data.push_back(carry);
    }
    return ret;
}

CBigInt* CBigInt::power(uint64_t pow) const {
    if (pow == 0)
        return new CBigInt(0);
    if (pow == 1)
        return new CBigInt(*this);

    if (!(pow % 2)) { // a^(2b) = 2(a^b)
        CBigInt* halfPower = power(pow / 2);
        CBigInt* ret = halfPower->mul(halfPower);
        delete halfPower;
        return ret;
    }

    CBigInt* ret = new CBigInt(*this);
    for (uint64_t i = 1; i < pow; ++i) {
        CBigInt* tmp = ret->mul(this);
        delete ret;
        ret = tmp;
    }

    ret->m_negative = (pow % 2 && m_negative);

    return ret;
}

bool CBigInt::addDestructive( CBigInt *oth ) {
    m_data.reserve( std::max(m_data.size() + 1, oth->m_data.size() + 1) );
    bool carry = 0;

    for (uint64_t i = 0; i < std::max(m_data.size(), oth->m_data.size()); ++i) {
        CHUNK_TYPE tmp = ( i < m_data.size() ? m_data[i] : 0 ) + ( i < oth->m_data.size() ? oth->m_data[i] : 0 ) + carry;
        carry = (tmp < ((i < m_data.size()) ? m_data[i] : oth->m_data[i]));
        if (m_data.size() > i)
            m_data[i] = tmp;
        else
            m_data.push_back(tmp);
    }
    if (carry)
        m_data.push_back(1);

    return true;
}

bool CBigInt::subDestructive( CBigInt *oth ) {
    bool borrow = 0;
    size_t len = oth->m_data.size();

    for (size_t i = 0; i < len; ++i) {
        CHUNK_TYPE tmp = m_data[i];
        if (borrow) {
            if (m_data[i] == 0) {
                tmp = CHUNK_MAX;
            } else {
                tmp = m_data[i] - 1;
                borrow = false;
            }
        }

        borrow = m_data[i] < oth->m_data[i];
        tmp -= oth->m_data[i];
        m_data[i] = tmp;
    }
    return true;
}

bool CBigInt::isZero( void ) const {
    if (!m_data.size())
        return true;
    
    for (uint64_t i = 0; i < m_data.size(); ++i)
        if (m_data[i])
            return false;
    
    return true;
}

std::pair<CBigInt*, CBigInt*> CBigInt::divmod( const CBigInt* oth) const {
    if (oth->isZero()) {
        throw std::logic_error("Division by zero");
    }

    std::unique_ptr<CBigInt> quotient( new CBigInt());
    std::unique_ptr<CBigInt> remainder( new CBigInt());

    const CBigInt *numerator(this);

    std::unique_ptr<CBigInt> divisor( new CBigInt(*oth));

    // Calculate the number of bits in the numerator and divisor
    int64_t numeratorBits = ((numerator->m_data.size() - 1) * CHUNK_BITS );
    if (numerator->m_data.size())
        numeratorBits += ( 64 - __builtin_clzll(numerator->m_data[numerator->m_data.size() - 1]) );
    int64_t divisorBits = ((divisor->m_data.size() - 1) * CHUNK_BITS);
    if (divisor->m_data.size())
        divisorBits += ( 64 - __builtin_clzll(divisor->m_data[divisor->m_data.size() - 1]) );
    
    int64_t shift = numeratorBits - divisorBits;

    if (shift < 0) {
        return { new CBigInt(), new CBigInt(*this) };
    }

    for (int64_t i = numeratorBits - 1; i >= 0; --i) {
        remainder->pushLeft( getBit(i) );

        if (remainder->cmp(divisor.get()) >= 0) {
            remainder->subDestructive(divisor.get());
            quotient->setBit(i, 1);
        }

    }

    quotient->m_negative = (m_negative ^ oth->m_negative);
    remainder->m_negative = false; // Ensure the remainder is positive

    quotient->optimize();
    remainder->optimize();

    return { quotient.release(), remainder.release() };
}

std::pair<CBigInt*, CBigInt*> CBigInt::rea( const CBigInt* oth) const {
//*       GCD       LCM
    std::unique_ptr<CBigInt> a(new CBigInt(*this));
    std::unique_ptr<CBigInt> b(new CBigInt(*oth));

    if (a->isZero() || b->isZero()) {
        return { a->add(b.get()), new CBigInt(0) };
    }

    while (!b->isZero()) {
        auto tmp = a->divmod(b.get());
        a.reset(b.release());
        b.reset(tmp.second);
        delete tmp.first;
    }

    std::unique_ptr<CBigInt> lcm(mul(oth));
    lcm.reset( lcm->div(a.get()) );

    return { a.release(), lcm.release() };
}

bool CBigInt::getBit( uint64_t i) const {
    if (i > m_data.size() * CHUNK_BITS)
        return false;
    
    return (m_data[ i / CHUNK_BITS ] & ( 1u << ( i % CHUNK_BITS )));
}

bool CBigInt::setBit( uint64_t i, bool b) {
    while (i >= m_data.size() * CHUNK_BITS) {
        m_data.push_back(0);
    }

    if (b)
        m_data[ i / CHUNK_BITS ] |= ( (CHUNK_TYPE)1 << (( i % CHUNK_BITS )));
    else
        m_data[ i / CHUNK_BITS ] &= ~( (CHUNK_TYPE)1 << (( i % CHUNK_BITS )));
    return true;
}

int CBigInt::cmp( const CBigInt *oth ) const { 
    if (m_data.size() != oth->m_data.size())
        return m_data.size() - oth->m_data.size();
    
    for (int64_t i = m_data.size() - 1; i >= 0; --i) {
        if (m_data[i] != oth->m_data[i])
            return 1 + ( -2 * (m_data[i] < oth->m_data[i]));
    }
    return 0;
}

bool CBigInt::pushLeft( bool i ) {
    bool carry = i;
    CHUNK_TYPE mask = (uint64_t)1 << (CHUNK_BITS - 1);
    for (uint64_t j = 0; j < m_data.size(); ++j) {
        bool tmp = m_data[j] & mask;
        m_data[j] <<= 1;
        m_data[j] |= carry;
        carry = tmp;
    }
    if (carry)
        m_data.push_back(1);
    
    return true;
}

bool CBigInt::optimize( void ) {
    if (m_data.size() < 1) {
        m_negative = false;
        return true;
    }
    // Delete trailing zeros
    for (int i = m_data.size() - 1; !m_data[i] && i ; m_data.pop_back(), --i);
    
    // Shrink the capacity if it's significantly larger than the size
    if (m_data.capacity() >= 2 * m_data.size())
        m_data.shrink_to_fit();

    // remove negative flag if its zero
    if (isZero())
        m_negative = false;

    return true;
} // delete zeros + shrink to fit