#pragma once
#include <vector>
#include <algorithm>
#include <iostream>
#include <memory>

#define CHUNK_TYPE uint32_t
#define CHUNK_BYTES sizeof(CHUNK_TYPE)
#define CHUNK_BITS (CHUNK_BYTES * 8)
#define CHUNK_MAX UINT32_MAX

/**
 * @class CBigInt
 * @brief Class representing a big integer.
 */
class CBigInt {
public:
    /**
     * @brief Default constructor.
     */
    CBigInt();

    /**
     * @brief Constructor that initializes the integer with a given value.
     * @param parent The value to initialize the integer with.
     */
    CBigInt( int64_t parent );

    /**
     * @brief Copy constructor.
     * @param oth The CBigInt object to copy.
     */
    CBigInt( const CBigInt &oth );

    /**
     * @brief Destructor.
     */
    ~CBigInt();

    /**
     * @brief Prints the big integer for debugging purposes.
     */
    void debugPrint();

    /**
     * @brief Dumps first 63bits and sign for debugging purposes.
     * @return The dumped value of the big integer.
     */
    int64_t debugDump();

    /**
     * @brief Converts the big integer to a string representation.
     * @return The string representation of the big integer.
     */
    std::string toString() const;

    /**
     * @brief Adds another big integer to this big integer.
     * @param oth The big integer to add.
     * @return A new big integer representing the sum.
     */
    CBigInt *add( const CBigInt *oth ) const;

    /**
     * @brief Subtracts another big integer from this big integer.
     * @param oth The big integer to subtract.
     * @return A new big integer representing the difference.
     */
    CBigInt *sub( const CBigInt *oth ) const;

    /**
     * @brief Multiplies another big integer with this big integer.
     * @param oth The big integer to multiply.
     * @return A new big integer representing the product.
     */
    CBigInt *mul( const CBigInt *oth ) const;

    /**
     * @brief Divides this big integer by another big integer.
     * uses divmod()
     * @param oth The divisor big integer.
     * @return A new big integer representing the quotient.
     */
    CBigInt *div( const CBigInt *oth ) const;

    /**
     * @brief Divides this big integer by another big integer.
     * uses divmod()
     * @param oth The divisor big integer.
     * @return A new big integer representing the quotient.
     */
    CBigInt *mod( const CBigInt *oth ) const;

    /**
     * @brief Calculates the greatest common divisor of the current big integer and another big integer.
     * uses rea()
     * @param oth The other big integer.
     * @return A new CBigInt object representing the greatest common divisor.
     */
    CBigInt *gcd( const CBigInt *oth ) const;

    /**
     * @brief Calculates the least common multiple of the current big integer and another big integer.
     * uses rea()
     * @param oth The other big integer.
     * @return A new CBigInt object representing the least common multiple.
     */
    CBigInt *lcm( const CBigInt *oth ) const;

    /**
     * @brief Performs a left shift operation on the current big integer.
     * @param shift The number of bits to shift.
     * @return A new CBigInt object representing the shifted big integer.
     */
    CBigInt *shf( int64_t shift ) const;

    /**
     * @brief Calculates the power of the current big integer raised to a given exponent.
     * @param pow The exponent.
     * @return A new CBigInt object representing the result of the power operation.
     */
    CBigInt *power( uint64_t pow ) const;

    /**
     * @brief Adds another big integer destructively to the current big integer.
     * @param oth The other big integer to be added.
     * @return A boolean indicating the success of the addition operation.
     */
    bool addDestructive( CBigInt *oth );

    /**
     * @brief Subtracts another big integer destructively from the current big integer.
     * @param oth The other big integer to be subtracted.
     * @return A boolean indicating the success of the subtraction operation.
     */
    bool subDestructive( CBigInt *oth );

    /**
     * @brief Pushes a bit to the left of the big integer.
     * @param i The bit to push to the left.
     * @return A boolean indicating the success of the push operation.
     */
    bool pushLeft( bool i );

    /**
     * @brief Gets the value of a specific bit at index i in the big integer.
     * @param i The index of the bit to retrieve.
     * @return The value of the bit at index i, or false if the index is out of range.
     */
    bool getBit( uint64_t i) const ;

    /**
     * @brief Sets the value of a specific bit at index i in the big integer.
     * @param i The index of the bit to set.
     * @param b The value to set the bit to (true or false).
     * @return A boolean indicating the success of the bit setting operation.
     */
    bool setBit( uint64_t i, bool b);

    /**
     * @brief Compares the current big integer with another big integer.
     * @param oth The other big integer to compare with.
     * @return An integer representing the comparison result: 0 if the numbers are equal, a negative value if the current number is smaller, and a positive value if the current number is larger.
     */
    int cmp( const CBigInt *oth ) const ;

    /**
     * @brief Optimizes the big integer by removing leading zero chunks.
     * @return A boolean indicating the success of the optimization operation.
     */
    bool optimize( void );

    /**
     * @brief Checks if the big integer is zero.
     * @return A boolean indicating whether the big integer is zero (true) or not (false).
     */
    bool isZero( void ) const ;

    /**
     * @brief Divides the current big integer by another big integer and calculates the remainder.
     * @param oth The divisor big integer.
     * @return A pair of CBigInt objects representing the quotient and remainder of the division.
     */
    std::pair<CBigInt *, CBigInt *> divmod( const CBigInt *oth ) const;

    /**
     * @brief Computes the GCD and LCM of two big integers.
     * @param oth The big integer to compute GCD and LCM with.
     * @return A pair of CBigInt pointers, representing the GCD and LCM respectively.
     */
    std::pair<CBigInt *, CBigInt *> rea( const CBigInt *oth ) const;

    /**
     * @brief Holds the chunks of the big integer's data.
     */
    std::vector<CHUNK_TYPE> m_data; 

    /**
     * @brief Indicates whether the big integer is negative.
     */
    bool m_negative;
};