#pragma once
#include "printable.hpp"
#include "bigint.hpp"
#include <cmath>

/**
 * @brief The CRational class represents the rational number in fraction format.
 * Responsible for controlling BigInts
 * It inherits from the CPrintable class.
 */
class CRational : public CPrintable {
public:
    /**
     * @brief Constructs a CRational object with a numerator and denominator.
     *
     * @param num Pointer to the numerator as a CBigInt object.
     * @param den Pointer to the denominator as a CBigInt object.
     */
    CRational( CBigInt *num, CBigInt *den);

    /**
     * @brief Constructs a CRational object with a numerator and a default denominator of 1.
     *
     * @param num Pointer to the numerator as a CBigInt object.
     */
    CRational( CBigInt *num );

    /**
     * @brief Constructs a CRational object as a copy of another CRational object.
     *
     * @param parent The CRational object to be copied.
     */
    CRational( const CRational &parent );

    /**
     * @brief Destroys the CRational object.
     */
    virtual ~CRational();

    /**
     * @brief Converts the CRational object to a string representation.
     */
    virtual CRational *evaluate() override;

    /**
     * @brief Converts the CRational object to a string representation.
     *
     */
    std::string toString();

    /**
     * @brief Simplifies the CRational object by dividing the numerator and denominator by their GCD.
     *        Adjusts the signs of the numerator and denominator accordingly to just nominator.
     */
    void simplify();

    /**
     * @brief Adds another CRational object to the current CRational object.
     *
     * @param oth Pointer to the CRational object to be added.
     * @return A new CRational object representing the sum of the two CRational objects.
     */
    CRational *add( CRational *oth );

    /**
     * @brief Subtracts another CRational object from the current CRational object.
     *
     * @param oth Pointer to the CRational object to be subtracted.
     * @return A new CRational object representing the difference between the two CRational objects.
     */
    CRational *sub( CRational *oth );

    /**
     * @brief Multiplies the current CRational object with another CRational object.
     *
     * @param oth Pointer to the CRational object to be multiplied.
     * @return A new CRational object representing the product of the two CRational objects.
     */
    CRational *mul( CRational *oth );

    /**
     * @brief Divides the current CRational object by another CRational object.
     *
     * @param oth Pointer to the CRational object to be divided by.
     * @return A new CRational object representing the division of the two CRational objects.
     */
    CRational *div( CRational *oth );

    /**
     * @brief Raises the current CRational object to a specified power.
     *
     * @param pow The exponent to raise the CRational object to.
     * @return A new CRational object representing the result of the exponentiation.
     */
    CRational *power( uint64_t pow );

    /**
     * @brief Computes a term of the CRational object raised to the specified power.
     * @param pow The power to raise the CRational object to.
     * @return A new CRational object representing the term raised to the specified power.
     */
    CRational *term( uint64_t pow );

    /**
     * @brief Computes the natural logarithm of the CRational object.
     * @note Throws an exception if the CRational object is non-positive.
     * @return A new CRational object representing the natural logarithm of the CRational object.
     */
    CRational *ln();

    /**
     * @brief Computes the exponential function of the CRational object.
     * @return A new CRational object representing the exponential function of the CRational object.
     */
    CRational *exp();

    std::string round( uint64_t decimal );

    /**
     * @brief Sets the same denominator for the current CRational object and another CRational object.
     * @param oth The other CRational object to set the same denominator with.
     * @return `true` if the operation is successful, `false` otherwise.
     */
    bool setSameDenominator( CRational *oth );

    /**
     * @brief Checks if the CRational number is negative.
     * @return `true` if the number is negative, `false` otherwise.
     */
    bool isNegative() const;

    /**
     * @brief Negates the CRational number by changing its sign.
     */
    void negate();

    /**
     * @brief Returns the debug dump of the CRational number as a double.
     * @return The debug dump value of the CRational number as a double.
     */
    double debugDump();
// private:

    /**
     * @brief Calculates the factorial of an unsigned integer.
     * @param i The unsigned integer to calculate the factorial for.
     * @return The factorial of the unsigned integer as a signed 64-bit integer.
     */
    inline static int64_t intfactorial( uint64_t i );

    /**
     * @brief The precision used for the term calculations.
     * It specifies the number of terms used in certain calculations.
     */
    static uint64_t termPrecision;

    /**
     * @brief The numerator of the CRational number.
     * It is stored as a unique pointer to a CBigInt object.
     */
    std::unique_ptr<CBigInt> m_numerator;

    /**
     * @brief The denominator of the CRational number.
     * It is stored as a unique pointer to a CBigInt object.
     */
    std::unique_ptr<CBigInt> m_denominator;
};